/*
	Virtual Lab Engine
	
	Lab scenes consist of "elements", many of which can be tabbed
	between.
		
*/

//	For the keyboard help audio.
var keyAudio = '../audio/GenericHelp.mp3';
//	For the keyboard help display.
var keyText = `
	<p>
		You can navigate this virtual lab using the keyboard.
	</p>
	<p>
		To advance through the lab step by step, press the <u>C</u> key.
	</p>
	<p>
		On certain scenes, you can press <u>I</u> for more information or
		<u>P</u> to see a photo.
	</p>
	<p>
		Press the <u>ESC</u> key to bring up the chapter menu. Press the
		<u>R</u> key to replay the audio on the current slide.
	</p>
	<p>
		Press the <u>T</u> key to toggle the audio.
	</p>
	<p>
		In general, when you see a button, the letter that is underlined on the
		button label will function as a keyboard command.  
	</p>
`;

var DARKENING_OVERLAY = `
				<div
					class="darkening-overlay"
					style="
					width: 800px;
					height: 600px;
					background-color: black;
					opacity: 0.25;
					pointer-events: none;
					transition: opacity 1s;
				">&nbsp;</div>
`;

var CLICK_BLOCKER = `
				<div
					id="click-blocker"
					style="
					width: 800px;
					height: 600px;
					background-color: transparent;
					opacity: 0.25;
					pointer-events: auto;
				">&nbsp;</div>
`;

var HOME_BUTTON = `
				<button style="
					left: 10px;
					top: 10px;
					font-size: 20px;
					text-align: center;
					border-radius: 20px;
					align-items: center;
					justify-content: center;
				"
				onclick="
					LAB.state.didHome();
				"
				><!-- 🏠 -->
					<div class="hamburger-menu">&#9776;</div>
					<div class="current-chapter"></div>
				</button>
`;

var DRAG_AREA = `
			<div id="drag-area"
				style="
					left: 0px;
					top: 0px;
					width: 800px;
					height: 800px;
					zbackground-color: rgba(255,0,0,0.3);
					pointer-events: auto;
					cursor: pointer;
				"
			>
				&nbsp;	
			</div>
`;

var MUTE_BUTTON = `
	<div id="MUTE-BUTTON"></div>
`;

/*
	Lab Class
	Acts as the main singleton of the lab.  Provides state management,
	narration services, and high-level keyboard support.

	narrate( url ) - Plays the given narration, stopping the previous
		narration if one is in progress.  Pass in null to halt all
		narration.  Narrations are PER-STATE, so that you can pause
		the narration for the current state, push a new state onto the
		stack which also plays a narration, pop it off the stack, and
		resume the original state's narration where it left off. This is
		useful for things like pop-up menus which should temporarily halt
		narration while you interact with them.
	
	pauseNarration() / resumeNarration() - Pauses or resumes narration.
	
	jumpTo( labState ) - Jump to the given state, ignoring previous stack.
	
	push( labState, callback ) - Add a state to the stack with a callback
		when that state is popped back off the stack.
	
	pop( returnValue ) - Pop a state off the stack and call the pushing
		state's callback with the given return value.
	
	switchTo( state ) - Switch to the given state within the current
		hierarchy depth value.  The new state becomes responsible for
		popping an expected return value compatible with the original
		state being switched from if it is not the first state in the
		stack.
	
	keyDown( e ) - Handles keydown events.  Sends the keydown to the
		current state.  (See keyDown(e) in the LabState class.)
		
	setCurrentChapter( # ) - Sets the current chapter number.
*/

class Lab {

	constructor() {
		this.currentChapter = 1;
		this.state = new LabState( { html: '' } );
		Lab.keyboardInteraction = false;
		this.keyListener = function(e) { this.keyDown(e); }.bind(this);
		document.addEventListener( 'keydown', this.keyListener );
		
		//	Watch for fullscreen changes.
		$(document).on(`
			webkitfullscreenchange
			mozfullscreenchange
			fullscreenchange
			MSFullscreenChange
			`, 
			function(e) {
				this.didChangeFullscreenStatus();
			}.bind(this)
		);
		
		this.bonkHowl = new Howl( {
			src: '../audio/UnavailableOption.mp3'
		} );
		
		this.controllerContainer = $(`
			<div style="
				position: absolute;
				left: 0px;
				top: 0px;"></div>
		`);
		$('#LAB').append( this.controllerContainer );
		$('#LAB').append( MUTE_BUTTON );
		$('#MUTE-BUTTON').click( function() {
			this.didClickMuteButton();
		}.bind(this));

	}
	
	didClickMuteButton() {
		Howler.mute( !Howler._muted );
		if (Howler._muted) {
			$('#MUTE-BUTTON').addClass( 'muted' );
		} else {
			$('#MUTE-BUTTON').removeClass( 'muted' );
		}
	}
	
	bonk() { this.bonkHowl.play(); }
	
	setCurrentChapter( chapter ) {
		this.currentChapter = chapter;
	}

	isFullscreen() {
		return this.fullscreenElement() != null;
	}
	
	fullscreenElement() {
		if (document.fullscreenElement != null) return document.fullscreenElement;
		if (document.mozFullscreenElement != null) return document.mozFullscreenElement;
		if (document.webkitFullscreenElement != null) return document.webkitFullscreenElement;
		if (document.msFullscreenElement != null) return document.msFullscreenElement;
		return null;
	}
	
	toggleFullscreen() {
	
		var lab = document.getElementById( 'LAB' );
		
		//	If it's fullscreen already, exit full screen.
		if (this.isFullscreen()) {
			//	Standard
			if (document.exitFullscreen) { document.exitFullscreen(); }
			//	Firefox
			else if (document.mozExitFullscreen) { document.mozExitFullscreen(); }
			//	Chrome, Safari, Opera, other webkit-based browsers
			else if (document.webkitExitFullscreen) { document.webkitExitFullscreen(); }
			//	IE / Edge
			else if (document.msExitFullscreen) { document.msExitFullscreen(); }
			return true;
		}
		
		//	Otherwise, enter full screen.
		//	Standard
		if (lab.requestFullscreen) { lab.requestFullscreen(); }
		//	Firefox
		else if (lab.mozRequestFullScreen) { lab.mozRequestFullScreen(); }
		//	Chrome, Safari, Opera, other webkit-based browsers
		else if (lab.webkitRequestFullscreen) { lab.webkitRequestFullscreen(); }
		//	IE / Edge
		else if (lab.msRequestFullscreen) { lab.msRequestFullscreen(); }
	}
	
	didChangeFullscreenStatus() {
		// document.fullscreenElement will point to the element that
		// is in fullscreen mode if there is one. If not, the value
		// of the property is null.
		var lab = document.getElementById('LAB');
		if (this.isFullscreen()) {
			var scale = Math.min( 
				window.screen.width / 800,
				window.screen.height / 600
			) * 0.95;
			$('#LAB').css( 'transform', 'scale(' + scale + ')' );
		} else {
			$('#LAB').css( 'transform', 'scale(1.0)' );
		}
	}
	
	pauseNarration() {
		if (this.state.narration == null) { return; }
		this.state.narration.pause();
	}
	
	resumeNarration() {
		if (this.state.narration == null) { return; }
		this.state.narration.play();
	}
	
	narrate( src ) {
		if (this.state.narration != null) {
			this.state.narration.stop();
			this.state.narration = null;
		}
		this.state.killKeyNotification();
		if (src == null) return;
		this.state.narration = new Howl( {
			src: src
		} );
		this.state.narration.play();
	}
	
	jumpTo( labState ) {
		this.waitFor( labState, function() {
			//	Quiet and depart all the states in the stack.
			while( this.state != null ) {
				this.narrate();
				if (this.state.depart) {
					this.state.depart();
				}
				this.state = this.state.parent;
			}
			this.controllerContainer.html('');
			this.state = labState;
			this.state.arrive();
		}.bind(this) );
	}
	
	push( state, callbackWhenPopped = null ) {
		let labState = state;
		let callback = callbackWhenPopped;
		this.waitFor( labState, function() {
			if (this.state == null) {
				this.state = labState;
				this.state.arrive();
				return;
			}
			if (labState.config.retainNarration == null) {
				this.pauseNarration();
			}
			if (this.state.root != null) {
				this.state.killInput();
			}
			//	Blur the button that sent us here so that the
			//	active element isn't a button.  This allows us
			//	to reliably call doContinue() when the user presses
			//	space and no button is focused in the child state.
			document.activeElement.blur();
			labState.parent = this.state;
			this.state = labState;
			this.state.callback = callback;
			this.state.arrive();
		}.bind(this) );
	}
	
	pop( returnValue = null ) {
		if (this.state == null) return;
		if (this.state.parent == null) return;
		this.narrate();
		this.state.depart();
		var oldState = this.state;
		this.state = this.state.parent;
		this.resumeNarration();
		this.state.restoreInput();
		if (oldState.callback) {
			oldState.callback( returnValue );
		}
	}
	
	switchTo( labState ) {
		if (this.state == null) {
			return this.push( labState );
		}
		this.waitFor( labState, function() {
			if (labState.config.retainNarration == null) {
				this.narrate();
			} else {
				labState.narration = this.state.narration;
			}
			let oldState = this.state;
			labState.parent = oldState.parent;
			labState.callback = oldState.callback;
			this.state = labState;
			oldState.depart();
			this.state.arrive();
		}.bind(this));
	}
	
	keyDown( e ) {
		if (!this.state) return false;
		if (!this.state.keyDown) return false;
		if (!this.state.active) return false;
		if (this.state.keyDown( e )) {
			e.preventDefault();
			return true;
		}
		if (e.key == "t") {
			this.didClickMuteButton();
			return true;
		}
		return false;
	}
	
	waitFor( labState, action ) {
		this.state.killInput();
		this.state.waitingAction = action;
		this.state.waitingForImages = [];
		var imageURLs = labState.images();
		for( var imageURL of imageURLs ) {
			var image = new Image();
			image.src = imageURL;
			if (image.complete) {
				continue;
			}
			this.state.waitingForImages.push( image );
		}
		requestAnimationFrame( function() { this.advanceOnImagesLoaded(); }.bind(this) );
	}
	
	advanceOnImagesLoaded() {
		if (this.state.waitingForImages == null) {
			if (this.state.waitingForAction) {
				this.state.restoreInput();
				this.state.waitingForAction();
			}
			return;
		}
		for( var i = this.state.waitingForImages.length - 1; i >= 0 ; i-- ) {
			var image = this.state.waitingForImages[i];
			if (image.complete) {
				this.state.waitingForImages.splice( i, 1 );
			}
		}
		if (this.state.waitingForImages.length == 0) {
			this.state.restoreInput();
			this.state.waitingAction();
			return;
		}
		requestAnimationFrame( function() { this.advanceOnImagesLoaded(); }.bind(this) );
	}
	
}


/*
	LabState class
	Responsible for representing a lab state.  Basically, it presents
	some HTML in the lab and handles moving to other states.
	
	Pass in a config object with the following elements:
	
		html -	The HTML for the scene.  Recommended to have at least
				one .balloon element.
		audio - (Optional) Audio to play when this slide is shown.
		onContinue - (Optional) A closure to perform when the user
				continues.  Adds a (C)ontinue button to the button bar.
				The button bar is either at the bottom of a .balloon
				div, or in the lower right of the screen if none found.
		onInfo - (Optional) As onContinue, but with (I)nfo.
		onPhoto - (Optional) As onContinue, but with (P)hoto.
		
	root - A jQuery object containing a "lab-state"-class div that will
		be pushed into the display area when this class is added to the
		state stack.  You can do whatever you need to do inside it to 
		present the lab state.  When it leaves the state stack, the root
		will be removed from the page.
	
	arrive() - Adds this state to the #LAB div.  Override to do more
		things when the LabState is added to the stack.  A good thing to
		do here is to create references to the various places you can go
		from here to start preloading them.
		
	depart() - Removes this state from the #LAB div.  Override to do more
		things when the LabState is removed from the stack.  A depart()
		call *should* be paired with every arrive() call.
		
	keyDown(e) - A keyDown event has been received while this state is
		at the top of the stack.  Return true to preventDefault() on the
		event (i.e., if this was handled).  By default, will attempt to
		call onContinue() on space, return, or enter if the current
		active element is not a button, or when typing a "c".  Pressing
		escape or "h" tries to call didHome().  Pressing "p" will try to
		call onPhoto() and pressing "i" will try to call onInfo().
	
	images() - Returns a naive list of images used by this lab state.
		This will be used for preloading the images.  If your state requires
		images that are not in the root DOM when initialization happens
		that need to be loaded for display, call images() in the
		superclass and add any other images needed:
		
			return [ ...super.images(), 'image1', 'image2' ];
	
	didHome() - Called when the user clicks the home button on a lab state.
		By default, will push a class called "PauseMenu" to the stack.
		Your particular lab should define this.  See existing labs as examples.
		
	blurParents() - Call to blur all this state's parents / ancestors.
	unblurParents() - Undo the blurParents() action.
	
	killInput() - Prevents all input for this state.
	restoreInput() - Restores after a killInput().
	allowClickThrough() - Allows a parent controller to remain clickable.
		
*/

class LabState {

	constructor( config ) {
		this.root = $('<div class="lab-state"></div>');
		this.config = config;
		if (config.html != null) {
			this.root.html( config.html );
		}
		if (config.collection != null) {
			this.root = config.collection;
		}
		this.active = true;
	}
	
	stateName() {
		var sname = '';
		if (this.constructor.name == 'LabState') return '';
		if (this.parent != null) sname = this.parent.stateName() + ' ';
		sname += this.constructor.name;
		return sname;
	}
	
	arrive() {
		$('#current-state').html( this.stateName() );
		if (this.config.chapter) {
			LAB.setCurrentChapter( this.config.chapter );
		}
		if (this.config.onPhoto) {
			this.addButton(
				'<b>P</b>hoto',
				'LAB.state.didPhoto();',
				'photo-button' );
		}
		if (this.config.onInfo) {
			this.addButton(
				'<b>I</b>nfo',
				'LAB.state.didInfo();',
				'info-button' );
		}
		if (this.config.onContinue) {
			this.addButton(
				'<b>C</b>ontinue',
				'LAB.state.didContinue();',
				'continue-button' );
		}
		if (this.config.audio != null) {
			var anchor = this.root.find('.balloon,.infobox');
			var indentX = -20;
			var indentY = -10;
			if (anchor.length == 0) {
				anchor = this.root; indentX = 40; indentY = 40;
			}
			anchor.append( `
			<div class="repeat-button-container" style="
				position:absolute;
				right: ${indentX}px;
				top: ${indentY}px;
				">
				<button style="
					width: 40px;
					height: 40px;
					padding: 0px;
					margin: 0px;
					border-radius: 20px;
					"
					onclick="LAB.narrate( LAB.state.config.audio );">
				<!-- 🔄  &#8634;-->
                <img src="../images/repeat.svg" alt=""/>
				</button>
			</div>
			`);
			if (this.config.retainNarration == null) {
				LAB.narrate( this.config.audio );
			}
		}
		
		LAB.controllerContainer.append( this.root );
		$('.current-chapter').html( 'Chapter<div class="chapter-number">' + LAB.currentChapter + '</div>' );
	}
	
	depart() {
		this.killKeyNotification();
		this.root.remove();
	}
	
	keyDown( e ) {
		if (!this.active) return false;
		if (this.keyNotifier != null) {
			if (e.key == ' ' || 
				e.key == 'Return' ||
				e.key == 'Enter' ||
				e.key == 'Escape' ||
				e.key == '/' ||
				e.key == '?') {
				this.killKeyNotification();
				return true;
			}
		}
		switch( e.key ) {
			case ' ':
			case 'Return':
			case 'Enter':
				if (document.activeElement.tagName != 'BUTTON') {
					this.didContinue();
					return true;
				}
				break;
			case 'Escape':
			case 'h':
				this.didHome();
				return true;
			case 'c':
				if (this.didContinue()) return true;
				LAB.bonk();
				break;
			case 'i':
				if (this.didInfo()) return true;
				LAB.bonk();
				break;
			case 'p':
				if (this.didPhoto()) return true;
				LAB.bonk();
				break;
			case 'r':
				if (this.config.audio) {
					LAB.narrate( this.config.audio );
					return true;
				}
				break;
			case '/':
			case '?':
				return this.notifyAboutKeys();
		}
		return false;
	}
	
	images() {
		var found = this.root.find('img').map( function( o, img ) { return img.src; } );
		if (this.parent != null) {
			found = { ...found, ...this.parent.images() };
		}
		return found;
	}
	
	didHome() {
		LAB.push( new PauseMenu() );
	}
	
	didContinue() {
		if (this.config.onContinue != null) {
			this.config.onContinue();
			return true;
		}
		return false;
	}
	
	didInfo() {
		if (this.config.onInfo != null) {
			this.config.onInfo();
			return true;
		}
		return false;
	}
	
	didPhoto() {
		if (this.config.onPhoto != null) {
			this.config.onPhoto();
			return true;
		}
		return false;
	}
	
	buttonBar() {
		//	Does one already exist?  If so, just return it.
		var buttonBar = this.root.find( '.button-bar' );
		if (buttonBar.length > 0) return buttonBar;
		//	One does not exist. Create it.
		buttonBar = $('<div class="button-bar"></div>');
		//	But where to put it?
		var balloon = this.root.find( '.balloon' );
		if (balloon.length > 0) {
			balloon.append( buttonBar );
		} else {
			this.root.append( buttonBar );
		}
		return buttonBar;
	}
		
	addButton( title, action, id = "button" ) {
		var buttonBar = this.buttonBar();
		buttonBar.append( `
				<button
					id="${id}"
					onclick="${action}"
				>
					${title}
				</button>
		`);
	}
	
	blur() {
		if (this.blurDepth == null) this.blurDepth = 0;
		this.blurDepth += 1;
		if (this.blurDepth >= 1) {
			this.root.css( 'filter', 'blur(5px)' );
		}
	}
	unblur() {
		if (this.blurDepth == null) return;
		this.blurDepth -= 1;
		if (this.blurDepth < 1) {
			this.root.css( 'filter', '' );
			delete( this.blurDepth );
		}
	}
	
	blurParents() {
		var state = this.parent;
		while( state != null ) {
			if (state.blur) {
				state.blur();
			}
			state = state.parent;
		}
		this.killKeyNotification();
	}
	
	unblurParents() {
		var state = this.parent;
		while( state != null ) {
			if (state.unblur) {
				state.unblur();
			}
			state = state.parent;
		}
	}
	
	killInput() {
		//	Remember I'm not active.
		this.active = false;
		//	Disable all the buttons.
		this.root.find('button').attr('disabled','disabled');
		//	Add the overlay.
		this.root.append( CLICK_BLOCKER );
		//	Kill a key notification if one exists.
		this.killKeyNotification();
	}
	
	restoreInput() {
		this.active = true;
		this.root.find('button').removeAttr('disabled');
		this.root.find('#click-blocker').remove();
	}
	
	allowClickThrough() {
		this.root.css('pointer-events', 'none');
		this.root.find( 'button' ).css( 'pointer-events', 'auto' );
		if (this.parent != null)
			this.parent.restoreInput();
	}
	
	notifyAboutKeys() {
	
		//	Check to see if we have a notifier.
		if (this.keyNotifier != null) {
			this.killKeyNotification();
			return true;
		}

		//	First, make sure we've set it up.
		if (keyAudio == null) return false;
		if (keyText == null) return false;
		
		//	Make the div.
		this.keyNotifier = $(`
			<div class="notification">
				${keyText}
			</div>
		`);
		
		//	If we're currently talking, pause narration.
		if (this.narration != null) {
			LAB.pauseNarration();
		}
		
		//	Show the notifier.
		this.root.append( this.keyNotifier );
		requestAnimationFrame( function() {
			this.keyNotifier.css('opacity',1).css('bottom', '-2px' );
		}.bind(this) );

		/*
		//	Play the audio.
		this.keyNotifier.narration = new Howl( {
			src: keyAudio
		} );
		this.keyNotifier.narration.play();

		setTimeout( function() {
			this.checkForNotificationDonePlaying();
		}.bind( this ), 100 );
		*/
		
		return true;
		
	}
	
	checkForNotificationDonePlaying() {
		if (this.keyNotificationStillPlaying()) {
			setTimeout( function() {
				this.checkForNotificationDonePlaying();
			}.bind( this ), 100 );
			return;
		}
		this.killKeyNotification();
	}
	
	killKeyNotification() {
		if (this.keyNotifier == null) return;
		if (this.keyNotifier.hasBeenKilled) return;
		//this.keyNotifier.narration.stop();
		this.keyNotifier.css('opacity',0).css('bottom','-100px' );
		setTimeout( function() {
			this.removeKeyNotification();
		}.bind(this), 500 );
		this.keyNotifier.hasBeenKilled = true;
	}
	
	removeKeyNotification() {
		if (this.keyNotifier == null) return;
		this.keyNotifier.remove();
		this.keyNotifier = null;
		LAB.resumeNarration();
	}
	
	keyNotificationStillPlaying() {
		if (this.keyNotifier == null) return false;
		if (this.keyNotifier.narration == null) return false;
		if (this.keyNotifier.narration.playing() == false) return false;
		if (this.keyNotifier.narration._src != keyAudio) return false;
		return true;
	}
	
}

/*
	Title menu class
	Provides a common, consistent way to present the opening placards.
	
	Pass in a config object with the following properties:
	
	placard		-	The URL to the placard.
	title		-	The lab title
	firstState	-	An instance of the first state.
	credits		-	A URL for credits (optional).
	
	Each lab should define its own TitleLabState class that inherits this.
*/

class TitleScreen extends LabState {

	constructor( config ) {
		var credits = '';
		if (config.credits != null) {
			credits = `
				<button
					onclick="LAB.state.didCredits();"
				>
					<b>A</b>bout this lab
				</button>
			`;
		}
		super({
			html:
				`
					<div class="placard-display">
						<div>
							<img src="${config.placard}" />
							<h1>
								${config.title}
							</h1>
							<button
								onclick="LAB.state.didContinue();"
								class="call-to-action"
							>
								<b style="font-weight: bold;">C</b>ontinue
							</button><br />
							<button
								onclick="LAB.state.didFullscreen();"
							>
								<b>F</b>ullscreen
							</button>
							${credits}
						</div>
					</div>
				`,
				audio: config.audio ?? null,
			} );
			this.firstState = config.firstState;
			this.credits = config.credits;
	}
	
	arrive() {
		super.arrive();
		//	We preload a silence sound, and play it when we
		//	continue to escape audio sandboxing features.
		this.silence = new Howl( {
			src: 'audio/Silence.mp3',
			preload: true,
			onloaderror: function() {
				console.warn(`
					Warning: could not preload audio/Silence.mp3 - this
					may cause problems with escaping the browser's audio
					sandboxing features.`);
			}
		} );
	}
	
	keyDown( e ) {
		switch( e.key ) {
			case 'a':
				if (this.credits != null) {
					this.didCredits();
				}
				break;
			case 'f':
				this.didFullscreen();
				return true;
		}
		return super.keyDown(e);
	}

	didContinue() {
		LAB.narrate( 'audio/Silence.mp3' );
		LAB.switchTo( this.firstState );
		return true;
	}
	
	didCredits() {
		if (this.credits != null) {
			LAB.pauseNarration();
			window.open(
				this.credits,
				"_blank");
		}
	}
	
	didFullscreen() {
		LAB.toggleFullscreen();
	}

}

/*
	End screen class
	Provides a common, consistent way to finish up the labs.
	
	Pass in a config object with the following properties:
	
	placard		-	The URL to the placard.
	title		-	The lab title
	firstState	-	An instance of the first state.
	
	Each lab should define its own EndLabState class that inherits this.
*/

class EndScreen extends LabState {

	constructor( config ) {
		if (config.moreInfo == null) config.moreInfo = ``;
		super({
			html:
				`
					<div class="placard-display">
						<div>
							<img src="${config.placard}" />
							<h1>The End</h1>
							<p>
								You have completed the 
								<em>${config.title}</em>
								virtual lab.
								<br />
								Visit the main page for more virtual
								labs you can explore.
							</p>
							${config.moreInfo}
							<button
								onclick="LAB.state.didContinue();"
							>
								<b>C</b>ontinue
							</button>
						</div>
					</div>
				`,
			audio: config.audio,
			} );
	}
	
	keyDown( e ) {
		switch( e.key ) {
			case 'f':
				this.didFullscreen();
				return true;
		}
		return super.keyDown(e);
	}

	didContinue() {
		LAB.switchTo( new TitleLabState() );
		return true;
	}
	
}

/*
	Pause menu class
	Provides a common, consistent way to present the pause menu.
	
	Pass in a config object with the following properties:
	
	chapters - A list of chapter objects with the following properties:
		label - The label on the button.
		key - Keyboard shortcut.
		action - A closure to call.
		
	Each lab should define its own PauseMenu class that inherits this.
	
*/
class PauseMenuState extends LabState {

	constructor( config ) {
	
		var options = '';
		for( var i = 0; i < config.chapters.length; i++ ) {
			var chapter = config.chapters[i];
			options += `
				<button
					onclick="LAB.state.goToChapter(${i});"
					><div class="pause-menu-button-label">${chapter.label}</div></button>
			`;
		}
	
		super( {
			...{
				html: `
					${DARKENING_OVERLAY}
					${DARKENING_OVERLAY}
					<div class="pause-menu">
						${options}
						<button
							id="continue"
							onclick="LAB.state.didContinue();"
							style="
								position: absolute;
								right: 25px;
								bottom: -25px;
							">
							<b>C</b>ontinue
						</button>
					</div>
					${HOME_BUTTON}
				` },
			...config
		});
	}
	
	arrive() {
		super.arrive();
		this.blurParents();
	}
	
	depart() {
		super.depart();
		this.unblurParents();
	}
	
	goToChapter( which ) {
		var chapter = this.config.chapters[which];
		chapter.action();
	}
	
	keyDown( e ) {
		for( var i = 0; i < this.config.chapters.length; i++ ) {
			var chapter = this.config.chapters[i];
			if (e.key == chapter.key) {
				this.goToChapter( i );
				return true;
			}
		}
		switch( e.key ) {
			case 'Escape':
			case 'c':
			case 'i':
			case 'h':
				this.didContinue();
				return true;
		}
		return super.keyDown(e);
	}

	didContinue() {
		LAB.pop();
	}

	didHome() {
		this.didContinue();
	}
	
}

//	Shows a picture of the lab tech speaking.  Pass in a config object
//	with the following properties:
//
//		html	-		HTML for what she says.
//		audio	-		The audio for what she says.
//		onContinue -	A closure for when the continue button is clicked.
//		onInfo -		A closure to call for an Info button (optional).
//						If info is not defined, no info button is shown.
//						Otherwise it is shown and activated with "I".
//		onPhoto -		Like onInfo, only with a Photo. Key command is "P".
//		backdrop - 		The backdrop image (optional).

class LabTechSpeaking extends LabState {

	constructor( config ) {
	
		if (config.backdrop == null) config.backdrop = 'images/1_Backdrop.png';
	
		//	Extend the configuration object with the HTML.
		let contentHTML = config.html;
		config = { ...config,
			...{
				html: `
				<img src="${config.backdrop}"
					style="
						filter: blur(5px);
						"
					/>
				${DARKENING_OVERLAY}
				<img src="images/1_Lab_Assistant.png"
					style="
						top: 100px;
						left: 50px;
						"
					/>
				<div class="balloon"
					style="
						left: 380px;
						top: 50px;
						width: 350px;
					"
				>
					<div class="dialogue">
						${contentHTML}
					</div>
				</div>
				${HOME_BUTTON}
			`}
		};
		super( config );
	}
	
	images() {
		return [ ...super.images(), '../images/balloon-portrait.png' ];
	}

}

//	Shows an info box.  Pass in a config object with the following properties:
//
//		html	-		HTML for the info box.
//		audio	-		The audio for the info box.

class InfoBoxState extends LabState {

	constructor( config ) {
		config = { ...config,
			...{
				html: `
				${DARKENING_OVERLAY}
				${DARKENING_OVERLAY}
				<div style="
					width: 800px;
					height: 600px;
					display: flex;
					align-items: center;
					justify-content: center;
				">
					<div
						class="infobox"
					>
						${config.html}
						<div class="button-bar">
						</div>
					</div>
				</div>
			`,
				onContinue: function() {
					LAB.pop();
				},
			}
		};
		super( config );
	}
	
	arrive() {
		super.arrive();
		this.blurParents();
	}

	depart() {
		super.depart();
		this.unblurParents();
	}

	keyDown( e ) {
		switch( e.key ) {
			case 'Escape':
				this.didContinue();
				return true;
		}
		return super.keyDown(e);
	}

}

//	Shows a photo box.  Pass in a config object with the following properties:
//
//		title	-		The title of the photo
//		image	-		URL for the image to display
//		caption	-		The caption for the photo
//		audio	-		If audio is available, the URL for it (optional).

class PhotoState extends InfoBoxState {
	constructor( config ) {
		super( {
			html: `
				<h2 style="text-align: center;">
					${config.title}
				</h2>

				<p style="text-align: center;">
					<img src="${config.image}" />
				</p>
					
				<p>
					${config.caption}
				</p>
			`,
			audio: config.audio
		} );
	}
}

//	Presents the user with a draggable activity.
//	Pass in a config object with the following properties:
//
//		html	-		The html of the scene.
//		audio	-		Audio to play.
//		drags	-		A list of DragAction objects to handle the drags.
//		onContinue - 	What to do when drags are complete.
//
//	You should also have a #drag-area div that can receive the mouse events.
//
//	Some functions to override:
//
//	startDrag( drag, x, y )
//		A function called when a drag or animation begins.
//	continueDrag( drag, x, y )
//		A function called while a drag continues.
//	endDrag( drag, x, y, didResolve )
//		A function called when a drag or animation ends.
		

class DragActivity extends LabState {

	startDrag( drag, x, y ) {}
	continueDrag( drag, x, y ) {}
	endDrag( drag, x, y, didResolve ) {}
	
	arrive() {
		super.arrive();
		window.requestAnimationFrame(
			function() {
				this.watchForClick();
			}.bind(this)
		);
		this.isAnimating = false;
		this.isDragging = false;
		this.allowClickThrough();
	}
	
	watchForClick() {
		var dragArea = document.getElementById('drag-area');
		if (dragArea == null) {
			console.warn("Could not find drag-area div to watch for drags on. Drags disabled.");
			return;
		}
		if (dragArea.length == 0) {
			console.warn("Could not find drag-area div to watch for drags on. Drags disabled.");
			return;
		}
		this.dragging = false;
		dragArea.addEventListener( 'mousedown', function( e ) {
			this.mouseDown( e );
		}.bind(this) );
		dragArea.addEventListener( 'mousemove', function( e ) {
			this.mouseMove( e );
		}.bind( this ) );
		dragArea.addEventListener( 'mouseup', function( e ) {
			this.mouseUp( e );
		}.bind( this ) );
		if (this.config.drags.length == 0) return;
		var drag = this.config.drags[0];
		drag.arrive();
		this.dragOnDeck( drag, false );
	}
	
	dragOnDeck( drag, isAnimating = false ) {}
	
	mouseDown( e ) {
		if (this.isAnimating) return;
		if (this.config.drags.length == 0) return;
		var drag = this.config.drags[0];
		this.isDragging = true;
		var thumb = $('#' + drag.thumbID).first();
		var currentX = parseInt( thumb.css('left' ));
		var currentY = parseInt( thumb.css('top' ));
		this.dragOffset = { x: currentX - e.clientX, y: currentY - e.clientY };
		this.startDrag( drag, currentX, currentY );
	}
	
	mouseMove( e ) {
		if (this.isAnimating) return;
		if (!this.isDragging) return;
		var drag = this.config.drags[0];
		if (drag.thumbID == null) {
			console.warn( 'Drag does not have thumbID.' );
			return;
		}
		var thumb = $('#' + drag.thumbID).first();
		if (thumb == null) {
			console.warn( 'Could not find thumb #' + drag.thumbID + '.');
			return;
		}
		var tryX = this.dragOffset.x + e.clientX;
		var tryY = this.dragOffset.y + e.clientY;
		var p = drag.constrain( { x: tryX, y: tryY } );
		thumb.css( 'left', p.x + 'px' );
		thumb.css( 'top', p.y + 'px' );
		var didResolve = drag.resolveTest( p );
		if ( didResolve ) {
			this.config.drags.shift();
			this.endDrag( drag, p.x, p.y, true );
			if (this.config.drags.length == 0) {
				this.isDragging = false;
				this.didContinue();
			} else if (drag.isDisjoint) {
				this.dragOnDeck( this.config.drags[0], false );
				this.isDragging = false;
			} else {
				this.dragOnDeck( this.config.drags[0], false );
				this.config.drags[0].arrive();
				this.startDrag( this.config.drags[0], p.x, p.y );
			}
			return;
		}
		this.continueDrag( drag, p.x, p.y );
	}
	
	mouseUp( e ) {
		if (this.isAnimating) return;
		if (!this.isDragging) return;
		if (this.config.drags.length == 0) return;
		var drag = this.config.drags[0];
		var thumb = $('#' + drag.thumbID).first();
		var currentX = parseInt( thumb.css('left' ));
		var currentY = parseInt( thumb.css('top' ));
		this.endDrag( drag, currentX, currentY, false );
		this.isDragging = false;
	}
	
	didContinue() {
		//	Start animation.
		if (this.isAnimating) return false;
		this.killInput();
		this.isAnimating = true;
		if (this.config.drags.length == 0) {
			this.config.onContinue();
			return true;
		}
		var drag = this.config.drags[0];
		
		drag.animate( function() {
			this.config.drags.shift();
			this.isAnimating = false;
			if (this.config.drags.length > 0) {
				this.dragOnDeck( this.config.drags[0], true );
			}
			this.didContinue();
		}.bind(this));
		return true;
	}
	
}

/*
	A drag action.  Pass in a configuration object with:
		thumbID
			The ID of the "thumb" element.
		isDisjoint
			Set this property to true if this "drag" is disjoint from
			the one following it (i.e., the next drag is "separate" and
			should not immediately flow with the current drag). Defaults
			to false, so that drags continue.  Can also set programmatically
			later (dragAction.isDisjoint = true).

	You can also override some of the following functions to provide
	specific functionality.
		
	arrive
		A function that is called when the drag comes "on deck".
	constrain
		A function that takes an {x,y} and returns an {x,y}
		to constrain the drag. By default, constrains to the screen.
	resolveTest
		A function that returns true if the { x,y } drag location
		should resolve, false if not. By default, it just always resolves.
	animate
		A function to animate the drag automatically.  Passes in
		a callback function to call when the animation is complete.
		By default, it skips the animation.  If the animation is completed,
		it should call the appropriate startDrag/continueDrag/endDrag
		functions in the state.

*/
class DragAction {
	constructor( config ) {
		this.config = config;
		this.thumbID = config.thumbID;
		this.isDisjoint = config.isDisjoint ? config.isDisjoint : false;
	}
	arrive() {}
	constrain( p ) {
		if (p.x < 0) return false;
		if (p.x > 800) return false;
		if (p.y < 0) return false;
		if (p.y > 600) return false;
		return true;
	}
	resolveTest( p ) {
		return true;
	}
	animate( callback ) {
		callback();
	}
	
}

/*
	Multipoint drag action.
		thumbID
			The ID of the "thumb" element.
		start
			A point where the drag starts {x:#,y:#}
		points
			A list of points. Resolves when all points touched. {x:#, y:#}
		radius
			How close you need to get to a point to resolve it.
		rect
			(Optional) A rectangle where the dragging is valid.
			{x:#, y:#, w:#, h:#}
		pointCallback
			(Optional) A callback for when a point is cleared.
		ms
			(Optional) Milliseconds between points when animating.
			Defaults to 500 (half a second).
		linear
			(Optional) Use linear interpolation between points when animating.
			Defaults to easing.
	
*/
class MultipointDragAction {
	constructor( config ) {
		this.config = config;
		this.thumbID = config.thumbID;
		if (config.radius == null) config.radius = 10;
		this.radiusSquared = config.radius * config.radius;
	}

	constrain( p ) {
		if (this.config.rect == null) return super.constrain( p );
		let rect = this.config.rect;
		var rp = { x: p.x, y: p.y };
		rp.x = Math.min( rect.x + rect.w, Math.max( rect.x, rp.x ) );
		rp.y = Math.min( rect.y + rect.h, Math.max( rect.y, rp.y ) );
		return rp;
	}
	resolveTest( p ) {
		//	Loop over all the points.
		var stillPointsRemaining = false;
		for( var i = this.config.points.length-1; i>= 0; i--) {
			var point = this.config.points[i];
			if (this.checkResolve( p, point )) {
				//	Resolved!
				//	Remove it from the list.
				this.config.points.splice( i, 1 );
				if (this.config.pointCallback != null) {
					this.config.pointCallback( point );
				}
			} else {
				stillPointsRemaining = true;
			}
		}
		return !stillPointsRemaining;
	}
	checkResolve( p1, p2 ) {
		let dx = p1.x - p2.x;
		let dy = p1.y - p2.y;
		let dSquared = dx * dx + dy * dy;
		return (dSquared <= this.radiusSquared);
	}
	
	arrive() {
		var thumb = $('#' + this.thumbID).first();
		thumb.css( 'left', this.config.start.x );
		thumb.css( 'top', this.config.start.y );
	}
	
	thumbLocation() {
		let thumb = $('#' + this.thumbID).first();
		let thumbX = parseInt( thumb.css('left') );
		let thumbY = parseInt( thumb.css('top') );
		return { x: thumbX, y: thumbY };
	}
	
	animate( callback ) {
		this.animateOutCallback = callback;
		let loc = this.thumbLocation();
		LAB.state.startDrag( this, loc.x, loc.y );
		this.isAnimating = true;
		this.animateToNextPoint();
		this.reportAnimationFrame();
	}
	
	//	Continuously reports continueDrag() for the thumb while
	//	isAnimating.
	reportAnimationFrame() {
		if (!this.isAnimating) return;
		requestAnimationFrame( function() {
			this.reportAnimationFrame();
		}.bind(this));
		let loc = this.thumbLocation();
		LAB.state.continueDrag( this,  loc.x, loc.y );
	}
	
	animateToNextPoint() {
		let easing = this.config.linear == true ? 'linear' : 'ease';
		let duration = this.config.ms != null ? this.config.ms : 500;
		let seconds = duration * 0.001;
		let loc = this.thumbLocation();
		if (this.config.points.length == 0) {
			this.isAnimating = false;
			LAB.state.endDrag( this, loc.x, loc.y );
			if (this.animateOutCallback != null) {
				this.animateOutCallback();
			}
			return;
		}
		let thumb = $('#' + this.thumbID).first();
		let nextPoint = this.config.points.shift();
		thumb.css( 'transition', 'all ' + seconds + 's ' + easing )
			.css('left', nextPoint.x + 'px' )
			.css('top', nextPoint.y + 'px');
		setTimeout( function() {
			if (this.config.pointCallback != null) {
				this.config.pointCallback( nextPoint );
			}
			this.animateToNextPoint();
		}.bind(this), duration);

	}
	
}

/*
	Point-to-point drag action.
		thumbID
			The ID of the "thumb" element.
		start
			A point where the drag starts {x:#,y:#}
		end
			A point where the drag ends {x:#,y:#}
		speed
			Speed to move the thumb when animating, in pixels per second.
*/
class PointToPointDragAction extends DragAction {

	constructor( config ) {
		super( config );
		this.a = this.config.start;
		this.b = this.config.end;
		this.delta = { x: this.b.x - this.a.x,
						y: this.b.y - this.a.y };
		this.d = Math.sqrt( this.delta.x * this.delta.x + this.delta.y * this.delta.y );
		this.unitVector = { x: this.delta.x / this.d, y: this.delta.y / this.d };
	}
	
	arrive() {
		var thumb = $('#' + this.thumbID).first();
		thumb.css( 'left', this.a.x );
		thumb.css( 'top', this.a.y );
	}
	
	findNearest( p, a, b ) {
		var atob = { x: b.x - a.x, y: b.y - a.y };
		var atop = { x: p.x - a.x, y: p.y - a.y };
		var len = atob.x * atob.x + atob.y * atob.y;
		var dot = atop.x * atob.x + atop.y * atob.y;
		var proportion = dot / len;
		var t = Math.min( 1, Math.max( 0, proportion ) );
		dot = ( b.x - a.x ) * ( p.y - a.y ) - ( b.y - a.y ) * ( p.x - a.x );
		return { x:a.x + atob.x * t, y: a.y + atob.y * t, t: t, done: t >= 0.999 };
	}

	constrain( p ) {
		//	project the point onto the line segment
		var nearest = this.findNearest( p, this.a, this.b );
		return nearest;
	}

	resolveTest( p ) {
		let c = this.constrain( p );
		return c.done;
	}
	
	animate( callback ) {
	
		var thumb = $('#' + this.thumbID).first();
		this.startX = parseInt( thumb.css('left') );
		this.startY = parseInt( thumb.css('top') );
		this.dx = this.config.end.x - this.startX;
		this.dy = this.config.end.y - this.startY;
		let dist = Math.sqrt( this.dx*this.dx + this.dy*this.dy );
		this.exitDuration = dist/ (this.config.speed * 10);

		this.callback = callback;
		this.startTime = performance.now();
		LAB.state.startDrag( this, this.startX,this.startY );
		this.continueAnimation( this.startTime );

	}
	
	continueAnimation( t ) {
		var thumb = $('#' + this.thumbID).first();
		//	Get the elapsed time.
		var elapsed = (t - this.startTime) * 0.001;
		//	Map it onto [0,1] based on the duration.
		var ramp = Math.max( 0, Math.min( 1, elapsed / this.exitDuration ));
		//	Ease.
		ramp = 1.0 - (Math.cos( ramp * Math.PI ) * 0.5 + 0.5);
		var cx = ramp * this.dx + this.startX;
		var cy = ramp * this.dy + this.startY;
		thumb.css( 'left', cx + 'px' );
		thumb.css( 'top', cy + 'px' );
		if (elapsed > this.exitDuration) {
			LAB.state.endDrag( this, cx, cy, true );
			this.callback();
		} else {
			LAB.state.continueDrag( this, cx, cy );
			requestAnimationFrame( function( t ) { this.continueAnimation( t ); }.bind(this) );
		}
	}

	OLDanimate( callback ) {
		var thumb = $('#' + this.thumbID).first();
		var cx = parseInt( thumb.css('left') );
		var cy = parseInt( thumb.css('top') );
		
		this.callback = callback;
		this.startTime = performance.now();
		this.callback = callback;
		this.continueAnimation( this.startTime );
		LAB.state.startDrag( this, cx,cy );
	}
	
	OLDcontinueAnimation( t ) {

		var elapsed = (t - this.startTime) * 0.001;
		
		var thumb = $('#' + this.thumbID).first();
		var cx = parseInt( thumb.css('left') );
		var cy = parseInt( thumb.css('top') );

		var nx = cx + this.unitVector.x * elapsed * this.config.speed;
		var ny = cy + this.unitVector.y * elapsed * this.config.speed;	
		
		var p = { x: nx, y: ny };	
		var constrained = this.constrain( p );
		
		thumb.css( 'left', constrained.x + 'px' );
		thumb.css( 'top', constrained.y + 'px' );
		if (constrained.done) {
			LAB.state.endDrag( this, constrained.x, constrained.y, true );
			this.callback();
		} else {
			LAB.state.continueDrag( this, constrained.x, constrained.y );
			requestAnimationFrame( function( t ) { this.continueAnimation( t ); }.bind(this) );
		}

	}
	
}

/*
	Horizontal drag action.
		thumbID
			The ID of the "thumb" element.
		y
			The y height to drag along.
		minX, maxX
			The draggable extents
		minResolveX, maxResolveX
			The extents of the area that resolves.
		speed
			Speed to move the thumb when animating, in pixels per second.
*/
class HorizontalDragAction extends DragAction {
	constructor( config ) {
		super( config );
	}

	constrain( p ) {

		p.x = Math.max( this.config.minX, Math.min( this.config.maxX, p.x ));
		p.y = this.config.y;
		return p;

	}

	resolveTest( p ) {

		if (p.x < this.config.minResolveX) return false;
		if (p.x > this.config.maxResolveX) return false;
		return true;

	}

	animate( callback ) {

		var thumb = $('#' + this.thumbID).first();
		this.animationStartX = parseInt( thumb.css('left') );
		this.destinationX = (this.config.minResolveX + this.config.maxResolveX) * 0.5;
		this.duration = Math.abs( this.destinationX - this.animationStartX ) / this.config.speed;
		this.startTime = performance.now();
		this.callback = callback;
		this.continueAnimation( this.startTime );
		LAB.state.startDrag( this, this.animationStartX, this.config.y );

	}
	
	continueAnimation( t ) {

		var elapsed = (t - this.startTime) * 0.001;
		var isDone = false;
		var proportion = elapsed / this.duration;
		if (proportion > 1.0) {
			proportion = 1.0;
			isDone = true;
		}
		var eased = Math.cos( proportion * Math.PI ) * -0.5 + 0.5;
		var pos = this.animationStartX + (this.destinationX - this.animationStartX) * eased;
		pos = parseInt( pos );
		var thumb = $('#' + this.thumbID).first();
		thumb.css( 'left', pos + 'px' );
		if (isDone) {
			LAB.state.endDrag( this, pos, this.config.y, true );
			this.callback();
		} else {
			LAB.state.continueDrag( this, pos, this.config.y );
			requestAnimationFrame( function( t ) { this.continueAnimation( t ); }.bind(this) );
		}

	}
	
}

/*
	Horizontal drag action.
		thumbID
			The ID of the "thumb" element.
		x
			The x column to drag along.
		minY, maxY
			The draggable extents
		minResolveY, maxResolveY
			The extents of the area that resolves.
		speed
			Speed to move the thumb when animating, in pixels per second.
*/
class VerticalDragAction extends DragAction {
	constructor( config ) {
		super( config );
	}

	constrain( p ) {
		p.y = Math.max( this.config.minY, Math.min( this.config.maxY, p.y ));
		p.x = this.config.x;
		return p;

	}

	resolveTest( p ) {
		if (p.y < this.config.minResolveY) return false;
		if (p.y > this.config.maxResolveY) return false;
		return true;
	}

	animate( callback ) {
		var thumb = $('#' + this.thumbID).first();
		this.animationStartY = parseInt( thumb.css('top') );
		this.destinationY = (this.config.minResolveY + this.config.maxResolveY) * 0.5;
		this.duration = Math.abs( this.destinationY - this.animationStartY ) / this.config.speed;
		this.startTime = performance.now();
		this.callback = callback;
		this.continueAnimation( this.startTime );
		LAB.state.startDrag( this, this.config.x, this.animationStartY );
	}
	
	continueAnimation( t ) {

		var elapsed = (t - this.startTime) * 0.001;
		var isDone = false;
		var proportion = elapsed / this.duration;
		if (proportion > 1.0) {
			proportion = 1.0;
			isDone = true;
		}
		var eased = Math.cos( proportion * Math.PI ) * -0.5 + 0.5;
		var pos = this.animationStartY + (this.destinationY - this.animationStartY) * eased;
		pos = parseInt( pos );
		var thumb = $('#' + this.thumbID).first();
		thumb.css( 'top', pos + 'px' );
		if (isDone) {
			LAB.state.endDrag( this, this.config.x, pos, true );
			this.callback();
		} else {
			LAB.state.continueDrag( this, this.config.x, pos );
			requestAnimationFrame( function( t ) { this.continueAnimation( t ); }.bind(this) );
		}

	}
	
}

/*
	Arbitrary drag action.
		thumbID
			The ID of the "thumb" element.
		dragRect
			A rect - x,y,width,height - determining draggable extents.
		resolveRect
			A rect determining resolvable extents.
		speed
			Speed to move the thumb when animating, in pixels per second.
		rise
			If defined, animating will rise up this many pixels as it
			travels.
*/
class TwoAxisDragAction extends DragAction {
	constructor( config ) {
		super( config );
		this.minX = config.dragRect.x;
		this.maxX = config.dragRect.x + config.dragRect.width;
		this.minY = config.dragRect.y;
		this.maxY = config.dragRect.y + config.dragRect.height;
		this.minResolveX = config.resolveRect.x;
		this.maxResolveX = config.resolveRect.x + config.resolveRect.width;
		this.minResolveY = config.resolveRect.y;
		this.maxResolveY = config.resolveRect.y + config.resolveRect.height;
	}

	constrain( p ) {
		p.x = Math.max( this.minX, Math.min( this.maxX, p.x ));
		p.y = Math.max( this.minY, Math.min( this.maxY, p.y ));
		return p;
	}

	resolveTest( p ) {
		if (p.x < this.minResolveX) return false;
		if (p.x > this.maxResolveX) return false;
		if (p.y < this.minResolveY) return false;
		if (p.y > this.maxResolveY) return false;
		return true;
	}

	animate( callback ) {

		var thumb = $('#' + this.thumbID).first();
		this.animationStartX = parseInt( thumb.css('left') );
		this.animationStartY = parseInt( thumb.css('top') );
		this.destinationX = (this.minResolveX + this.maxResolveX) * 0.5;
		this.destinationY = (this.minResolveY + this.maxResolveY) * 0.5;
		var dx = this.destinationX - this.animationStartX;
		var dy = this.destinationY - this.animationStartY;
		var d = Math.sqrt( dx * dx + dy * dy );
		this.duration = d / this.config.speed;
		this.startTime = performance.now();
		this.callback = callback;
		this.continueAnimation( this.startTime );
		LAB.state.startDrag( this, this.animationStartX, this.animationStartY );

	}
	
	continueAnimation( t ) {

		var elapsed = (t - this.startTime) * 0.001;
		var isDone = false;
		var proportion = elapsed / this.duration;
		if (proportion > 1.0) {
			proportion = 1.0;
			isDone = true;
		}
		var rise = this.config.rise != null ? this.config.rise : 0;
		rise = rise * Math.sin( proportion * Math.PI );
		var eased = Math.cos( proportion * Math.PI ) * -0.5 + 0.5;
		var posX = this.animationStartX + (this.destinationX - this.animationStartX) * eased;
		var posY = this.animationStartY + (this.destinationY - this.animationStartY) * eased - rise;
		posX = parseInt( posX );
		posY = parseInt( posY );
		var thumb = $('#' + this.thumbID).first();
		thumb.css( 'left', posX + 'px' );
		thumb.css( 'top', posY + 'px' );
		if (isDone) {
			LAB.state.endDrag( this, posX, posY, true );
			this.callback();
		} else {
			LAB.state.continueDrag( this, posX, posY );
			requestAnimationFrame( function( t ) { this.continueAnimation( t ); }.bind(this) );
		}

	}
	
}

/*
	Arbitrary drag action.
		thumbID
			The ID of the "thumb" element.
		target
			An {x,y,r} object describing a location and a 
			radius.
		speed
			Speed to move the thumb when animating, in pixels per second.
*/
class DragToTarget extends TwoAxisDragAction {

	constructor( config ) {
		var r = config.target.r;
		config = {
			...{
				dragRect: { x: r, y: r, width: 800-r, height: 600-r },
				resolveRect: {
					x: config.target.x - r,
					y: config.target.y - r,
					width: 2*r, height: 2*r
				},
				speed: 300,
				rise: 50,
			},
			...config
		};
		super( config );
		
	}

}

//	Presents the user with an activity where they turn a spindle.
//	Pass in a config object with the following properties:
//
//		html		-	The html of the scene.
//		audio		-	Audio to play.
//		pivot		-	The ID of the div to spin.
//		startAngle	-	The starting angle, in degrees.
//		minAngle	-	The minimum angle, in degrees.
//		maxAngle	-	The maximum angle, in degrees.
//		targetAngle	-	The angle we want the user to drag to.
//		epsilon		-	The epsilon for the targetAngle. (Defaults to 1°.
//						If the target is not a min or max, you probably
//						want a much larger epsilon.
//		onContinue 	- 	What to do when drags are complete.
//		rotateSpeed	-	How fast to rotate when animating out.
//
//	You should also have a #drag-area div that can receive the mouse events.
//
//	Some functions to override:
//
//	startRotation( angle )
//		A function called when a drag or animation begins.
//	continueRotation( angle )
//		A function called while a drag continues.
//	endRotation( angle, didResolve )
//		A function called when a drag or animation ends.
		

class SpinActivity extends LabState {

	constructor( config ) {
		var whenDone = config.onContinue;
		config.onContinue = function() {
			LAB.state.didResolve();
		};
		super( config );
		this.whenDone = whenDone;
		this.startAngle = config.startAngle;
		this.minAngle = config.minAngle;
		this.maxAngle = config.maxAngle;
		this.targetAngle = config.targetAngle;
		this.epsilon = config.epsilon == null ? 3 : config.epsilon;
		this.display = this.root.find('#' + config.pivot);
		this.rotateSpeed = config.rotateSpeed == null ? 90 : config.rotateSpeed;
	}

	startRotation( angle ) {}
	continueRotation( angle ) {}
	endRotation( angle, didResolve ) {}
	
	arrive() {
		super.arrive();
		window.requestAnimationFrame(
			function() {
				this.watchForClick();
			}.bind(this)
		);
		this.setAngle( this.config.startAngle );
		this.isAnimating = false;
		this.isDragging = false;
		this.allowClickThrough();
	}
	
	setAngle( ang ) {
		this.display.css('transform','rotate(' + ang + 'deg)');
		this.angle = ang;
	}
	
	//	This gets the angle of the point relative to the pivot.
	//	Since we may want to have multiple spins, we return the
	//	angle "nearest" to a given angle.  For instance, is we click
	//	at 10° but we want the angle "nearest" to 360°, it will
	//	return 370°, not 10°. This way, we can enforce a number of
	//	rotations, asking the user to rotate towards 720°, for instance.
	getAngle( p, nearAngle = 0 ) {
	    var angle = Math.atan2(p.y, p.x);
	    var degrees = 180 * angle / Math.PI;
	    while( degrees > nearAngle - 360 ) {
	    	degrees -= 360;
	    }
	    var nearest = degrees;
	    var distance = Math.abs( nearAngle - degrees );
	    while( degrees < nearAngle + 360 ) {
	    	degrees += 360;
	    	var d = Math.abs( nearAngle - degrees );
	    	if (d < distance) {
	    		nearest = degrees;
	    		distance = d;
	    	}
	    }
	    return nearest;
	}
	
	watchForClick() {
		var dragArea = document.getElementById('drag-area');
		if (dragArea == null) {
			console.warn("Could not find drag-area div to watch for drags on. Drags disabled.");
			return;
		}
		if (dragArea.length == 0) {
			console.warn("Could not find drag-area div to watch for drags on. Drags disabled.");
			return;
		}
		this.dragging = false;
		dragArea.addEventListener( 'mousedown', function( e ) {
			this.mouseDown( e );
		}.bind(this) );
		dragArea.addEventListener( 'mousemove', function( e ) {
			this.mouseMove( e );
		}.bind( this ) );
		dragArea.addEventListener( 'mouseup', function( e ) {
			this.mouseUp( e );
		}.bind( this ) );
	}
	
	mousePositionRelativeToSpindle( e ) {
		var clickX = e.pageX - this.display.offset().left;
		var clickY = e.pageY - this.display.offset().top;
		return { x: clickX, y: clickY };
	}
	
	//	When we mouse down, remember the angle we started at relative to
	//	the current angle.  For instance, if I click at 30° when the spindle
	//	is at 0°, then when I drag to 40°, the spindle should be at 10°
	//	because I would have rotated 10° around the spindle.  In this case,
	//	this.angleOffset would be 30, because it is 30° "ahead" of the spindle
	//	while we drag.
	mouseDown( e ) {
		if (this.isAnimating) return;
		this.isDragging = true;
		var p = this.mousePositionRelativeToSpindle( e );
		this.angleOffset = this.getAngle( {
			x: p.x, y:p.y } ) - this.angle;
		this.startRotation( this.angle );
	}
	
	mouseMove( e ) {
		if (this.isAnimating) return;
		if (!this.isDragging) return;
		var p = this.mousePositionRelativeToSpindle( e );
		var dragAngle = this.getAngle( { x: p.x,y: p.y }, this.angle + this.angleOffset );
		dragAngle = dragAngle - this.angleOffset;
	    dragAngle = Math.max( this.minAngle, Math.min( this.maxAngle, dragAngle ));
		this.setAngle( dragAngle );
		this.continueRotation( dragAngle );
		//	Check for resolve.
		if (Math.abs( this.angle - this.targetAngle ) < this.epsilon ) {
			this.setAngle( this.targetAngle );
			this.endRotation( this.angle, true );
			this.isDragging = false;
			this.killInput();
			this.whenDone();
			return;
		}
	}
	
	mouseUp( e ) {
		if (this.isAnimating) return;
		if (!this.isDragging) return;
		this.endRotation( this.angle, false );
		this.isDragging = false;
	}
	
	didResolve() {
		//	Animate out.
		this.killInput();
		this.isAnimating = true;
		this.lastFrame = performance.now();
		requestAnimationFrame( function( timestamp ) {
			this.animateOut( timestamp ); }.bind(this) );
		this.startRotation( this.angle );
		//	TODO : Don't emit updates if we resolved with an end.
	}
	
	animateOut( timestamp ) {
		var delta = (timestamp - this.lastFrame) * 0.001;
		this.lastFrame = timestamp;
		var didFinish = false;
		if (this.angle < this.targetAngle) {
			this.angle += delta * this.rotateSpeed;
			if (this.angle > this.targetAngle) {
				didFinish = true;
				this.angle = this.targetAngle;
			}
		} else {
			this.angle -= delta * this.rotateSpeed;
			if (this.angle < this.targetAngle) {
				didFinish = true;
				this.angle = this.targetAngle;
			}
		}
		this.setAngle( this.angle );
		if (didFinish) {
			this.endRotation( this.angle );
			this.whenDone();
		} else {
			this.continueRotation( this.angle );
			requestAnimationFrame( function( timestamp ) {
				this.animateOut( timestamp ); }.bind(this) );
		}
	}
	
}

//	Helper function for randomizing stuff.
function d(sides) {
	return Math.floor(Math.random() * sides) + 1;
}

/*
	An object that manages a flipbook.  A flipbook must consist of a div
	that has the frames as the children in the order you want them displayed
	in.  Use CSS to position the container and the individual images however
	you like. Pass the container div in the constructor.
	
	Instead of typing out all the frames or adding them programmatically in
	the slide, or defining the playback speed programmatically, you can
	define some attributes on the flipbook element:
	
		<div
			id="my-flipbook"
			class="flipbook"
			ticksperframe="1000"
			framecount="12"
			frameprefix="images/my-flipbook-"
			framesuffix=".png"></div>
	
	This will automatically load images/my-flipbook-1.png through
	images/my-flipbook-12.png and add them to the flipbook automatically.
	
	You can: playOnce( ticks/frame ) or playForever( ticks/frame ) or stop().
	
	The ticks/frame defaults to 30 fps.
	
*/
		
class FlipBook {

	constructor( container ) {
		this.container = container;
		this.loopMode = false;
		if (this.container.attr('framecount') != null) {
			this.loadFrames(
				this.container.attr('frameprefix'),
				this.container.attr('framesuffix'),
				parseInt( this.container.attr('framecount'))
			);
		}
		this.ticksPerFrame = 33;	//	Defaults to 30fps
		if (this.container.attr('ticksperframe')) {
			this.ticksPerFrame = parseInt( this.container.attr('ticksperframe'));
		}
		this.frames = this.container.find('img').length;
		this.showFrame( 1 );
	}
	
	//	Load a series of frames.
	//	You can pass in an array of src attributes OR
	//	you can pass in a prefix, a suffix, and a number of frames.
	//	E.g., if you pass in ('images/thing_', '.png', 2), it will
	//	load "images/thing_1.png" and "images/thing_2.png".
	loadFrames( prefix, suffix, numberOfFrames ) {
		if (Array.isArray( prefix ) == true) {
			for (var frame of prefix ) {
				this.container.append( '<img src="' + frame + '"/>' );
			}
			this.frames = this.container.find('img').length;
			return;
		}
		for( var i = 1; i <= numberOfFrames; i++ ) {
			this.container.append( '<img src="' + prefix + i + suffix + '"/>' );
		}
		this.frames = this.container.find('img').length;
	}
	
	loadFrame( src ) {
		this.container.append( '<img src="' + src + '"/>' );
		this.frames = this.container.find('img').length;
	}
	
	stop() {
		if (this.interval == null) return;
		clearInterval( this.interval );
		this.interval = null;
	}
	
	playOnce( ticksPerFrame ) {
		if (ticksPerFrame == null) ticksPerFrame = this.ticksPerFrame;
		if (this.interval != null) return;
		this.loopMode = false;
		this.interval = setInterval( function() {
			this.advance();
		}.bind(this), ticksPerFrame );
	}
	
	playOnceBackwards( ticksPerFrame ) {
		if (ticksPerFrame == null) ticksPerFrame = this.ticksPerFrame;
		if (this.interval != null) return;
		this.loopMode = false;
		this.interval = setInterval( function() {
			this.reverse();
		}.bind(this), ticksPerFrame );
	}
	
	playForever( ticksPerFrame = 33 ) {
		if (this.interval != null) return;
		this.loopMode = true;
		this.interval = setInterval( function() {
			this.advance();
		}.bind(this), ticksPerFrame );
	}
	
	showFrame( frame ) {
		this.frame = frame;
		this.container.find('img').css('opacity', '0');
		this.container.find('img:nth-child(' + frame + ')').css('opacity', '1');
	}
	
	advance() {
		//	If we've left the page, auto-stop.
		if ($('#LAB').find( this.container ).length == 0) {
			this.stop();
			return;
		}
		this.frame += 1;
		if (this.frame > this.frames) {
			//	We're done.  If we're looping, loop.
			if (this.loopMode) {
				this.showFrame(1);
				return;
			} else {
				this.showFrame( this.frames );
				clearInterval( this.interval );
				this.interval = null;
				return;
			}
		}
		this.showFrame( this.frame );
	}
	
	reverse() {
		//	If we've left the page, auto-stop.
		if ($('#LAB').find( this.container ).length == 0) {
			this.stop();
			return;
		}
		this.frame -= 1;
		if (this.frame < 1) {
			//	We're done.  If we're looping, loop.
			if (this.loopMode) {
				this.showFrame(this.frames);
				return;
			} else {
				this.showFrame( 1 );
				clearInterval( this.interval );
				this.interval = null;
				return;
			}
		}
		this.showFrame( this.frame );
	}
	
}

